module Numerical

using LinearAlgebra
using AbstractAlgebra
using HomotopyContinuation
using ..Algebraic

# --- polynomial conversion ---

# hat tip Sascha Timme
#   https://github.com/JuliaHomotopyContinuation/HomotopyContinuation.jl/issues/520#issuecomment-1317681521
function Base.convert(::Type{Expression}, f::MPolyRingElem)
  variables = Variable.(symbols(parent(f)))
  f_data = zip(AbstractAlgebra.coefficients(f), exponent_vectors(f))
  sum(cf * prod(variables .^ exp_vec) for (cf, exp_vec) in f_data)
end

# create a ModelKit.System from an ideal in a multivariate polynomial ring. the
# variable ordering is taken from the polynomial ring
function System(I::Generic.Ideal)
  eqns = Expression.(gens(I))
  variables = Variable.(symbols(base_ring(I)))
  System(eqns, variables = variables)
end

# --- sampling ---

function real_samples(F::AbstractSystem, dim)
  # choose a random real hyperplane of codimension `dim` by intersecting
  # hyperplanes whose normal vectors are uniformly distributed over the unit
  # sphere
  # [to do] guard against the unlikely event that one of the normals is zero
  normals = transpose(hcat(
    (normalize(randn(nvariables(F))) for _ in 1:dim)...
  ))
  cut = LinearSubspace(normals, fill(0., dim))
  filter(isreal, results(witness_set(F, cut)))
end

end